'********************************************************************
' PBVLIST32, Virtual ListBox custom control for PB/DLL
'********************************************************************
' Public Domain by Borje Hagsten, December 2000
'
' LOG:
' May  4, 2003: Changed behaviour under %WM_GETDLGCODE to let system
'                 handle VK_TAB, because GetNextDlgTabItem does not
'                 seem to work as expected in some situations.
'               Added DrawFocusRect under WM_SETFOCUS, WM_KILLFOCUS
'                 and WM_PAINT for situations when control has focus,
'                 but nothing is selelected, to mimic standard listbox's
'                 behaviour there.
'               Added search for next item that starts same as pressed
'                 key under WM_CHAR, to mimic standard Listbox's behaviour.
'               Replaced TextOut with TabbedTextOut in WM_PAINT to enable
'                 tab-delimited text. Tab stops are set to default, so
'                 you need to change the 3 last parameters to change this.
'                 See Win32.hlp for how to do this. Note: If text contains
'                 no tab stops, it works like normal TextOut.
'               Made some minor changes to creation code to prepare for DLL
'                 compile and replaced LocalAlloc with HeapAlloc in WM_CREATE.
'
' Apr 15, 2003: Added %VLB_SETCURSEL message that mimics LB_SETCURSEL
'                 in that it both sets sel and makes sure selected item
'                 is scrolled into view (if ourside).
'               Added %VLB_GETCURSEL to match the above. It is so far
'                 the same as %VLB_GETSELECTED, but possibly the later
'                 will be expanded for multiple sel some day, so..
'
' Jan 14, 2003: Changed to DWORD for handles in some places and
'                 now use GetWindowLong(hWnd, %GWL_HINSTANCE) to
'                 get proper instance handle at creation.
'               Added lParam check to %WM_GETDLGCODE in VListProc.
'               Removed unnecessary FUNCTION = 0 in VListProc.
' Nov 30  2002: Minor adjustments for latest Win32api.inc

' Save this file as PBVLIST.INC in for example your Pbwin70\Winapi
' folder, and then declare #INCLUDE "PBVLIST.INC" in code to enable use.
'
' Needed a way to view huge arrays very fast. This is the result.
' Free to use and modify to suit your own needs. Commented code, should
' be "easy" to follow. Tested for memory leaks and sprayed for bugs.
' Still, as always - use at own responsibility. Introduction is a bit
' long, but includes both manual and some explanations.
'
' Special thanks to:
'   Chris Boss, for all valuable tips about custom controls, etc.
'   Jules Marchildon, for his "Supercharged Repaints" sample, which
'   lead me to find a flicker-free, lean and fast way to draw the list.
'   Lance Edmunds, Dave Navarro and the entire PB community for all hints
'   and tips, always making it worth while to visit the PB forums.. :-)
'
'********************************************************************
' Manual
'********************************************************************
' Basic version, for instant display of array of any size in one column,
'                vertical scroll bar only. Modify if you have other needs.
'
' Advantages:    Lean and mean. Uses existing data and only visible items
'                are drawn, which makes it very fast and lean on memory.
'                Good for viewing huge arrays.
'
' Disadvantages: Only works with global arrays.
'                No direct editing of Listbox items, but via PB's ARRAY
'                functions, this is easily solved. Sample code shows how.
'
' File(s)        PBVLIST.INC, include file for the Virtual ListBox
'                PBVLIST.BAS, example program for the Virtual ListBox
'
' This Virtual Listbox sort of works like a "Visual Array", so the
' array to be shown must be global. Any changes to list items must be
' done by working directly with the array and then use %VLB_SETARRAY
' and %VLB_REFRESH messages to refresh the list. See sample code in
' PBVLIST.BAS. PowerBasic has some excellent tools for working with
' arrays via ARRAY SORT, ARRAY SCAN, ARRAY DELETE and ARRAY INSERT.
'
' Selection is drawn with default colors when control has focus and grayed
' when it loses focus, for a better view of which control that has focus.
' Code for this is in WM_PAINT, if you prefer standard behaviour instead.
'
' Control sends %WM_CTLCOLORLISTBOX in %WM_PAINT to parent, to get handle
' of background color brush (if the message is handled by parent dialog).
'
' Creation can be made either DDT or SDK way, in CreateVirtualList(..)
'
' On font change, control resizes itself to to avoid partial lines (see WM_SETFONT).
'
' Additional notification messages added for easier handling of
' Enter, Space bar and Delete keys.
'
' Tip: Items are drawn in WM_PAINT. You can easily change things there to
'      suit you own needs. For columns, you can use TabbedTextOut, or parse
'      each item and use multiple TextOut's with x set to proper values for
'      each column. Using SetTextColor, you can even set different colors to
'      each column or word, etc. Instead of using an array, one could also
'      connect it to a database file, where visible items are read directly
'      from the file, in WM_PAINT. With only a few changes and the addition
'      of a caret, you can even turn it into an edit control.. :-)
'
'--------------------------------------------------------------------
' Custom-made messages:
'--------------------------------------------------------------------
' %VLB_SETARRAY     - to set the global array to be shown in the listbox
' wParam = A pointer to first element to show in array. Can be anything
'          between LBOUND(Array) to UBOUND(Array).
' lParam = The number of elements to show. Typically UBOUND(Array), but can be
'          any positive value up to UBOUND(Array).
' Returns: Nothing, 0
'
' Example: To show entire array:
'             SendMessage hWndCtrl, %VLB_SETARRAY, VARPTR(array(LBOUND(array))), UBOUND(array)
'             SendMessage hWndCtrl, %VLB_REFRESH, 0, 0
'          To show elements 10 to 50:
'             SendMessage hWndCtrl, %VLB_SETARRAY, VARPTR(array(10)), 40
'             SendMessage hWndCtrl, %VLB_REFRESH, 0, 0
' Tip:     Use this message in combination with %VLB_SETTOPLINE, %VLB_SETSELECTED
'          and %VLB_REFRESH to refresh list after changes to the array.
'--------------------------------------------------------------------
' %VLB_REFRESH      - to refresh control with new array data
' wParam = Not used, should be 0.
' lParam = Not used, should be 0.
' Returns: Nothing, 0
' Example: SendMessage hWndCtrl, %VLB_REFRESH, 0, 0
'--------------------------------------------------------------------
' %VLB_SIZEHANDLER  - internal, for handling resizing and memory DC's etc.
' wParam = Not used, should be 0.
' lParam = Not used, should be 0.
' Returns: Nothing, 0
' Example: SendMessage hWndCtrl, %VLB_SIZEHANDLER, 0, 0
'--------------------------------------------------------------------
' %VLB_GETSELECTED  - returns selected line number ((array item)
' wParam = not used, should be 0.
' lParam = not used, should be 0.
' Returns: Index of selected line.
' Example: index = SendMessage(hWndCtrl, %VLB_GETSELECTED, 0, 0)
'--------------------------------------------------------------------
' %VLB_SETSELECTED  - sets selected line number ((array item)
' wParam = Line to select.
' lParam = Redraw flag. %TRUE to redraw, %FALSE to ignore
' Returns: Index of selected line, or -1 if no selection could be made.
' Example: index = SendMessage(hWndCtrl, %VLB_SETSELECTED, SelectLine, %TRUE)
'--------------------------------------------------------------------
' %VLB_GETTOPLINE   - returns first visible line in control
' wParam = not used, should be 0.
' lParam = not used, should be 0.
' Returns: Index of control's top (first visible) line.
' Example: index = SendMessage(hWndCtrl, %VLB_GETTOPLINE, 0, 0)
'--------------------------------------------------------------------
' %VLB_SETTOPLINE   - sets first visible line in control
' wParam = Line to show as top (first visible) line.
' lParam = Redraw flag. %TRUE to redraw, %FALSE to ignore
' Returns: Index of top line, or -1 if no items are available.
' Example: index = SendMessage(hWndCtrl, %VLB_SETTOPLINE, TopLine, %TRUE)
'--------------------------------------------------------------------
' %VLBN_RETURN
' Notification sent to parent's %WM_COMMAND when Enter key has been pressed
'--------------------------------------------------------------------
' %VLBN_SPACE
' Notification sent to parent's %WM_COMMAND when Space bar has been pressed
'--------------------------------------------------------------------
' %VLBN_DELETE
' Notification sent to parent's %WM_COMMAND when Delete key has been pressed

'********************************************************************
' Declares
'********************************************************************
DECLARE SUB SetPBVLScrollData(BYVAL hwnd AS DWORD, BYVAL vhPage AS LONG, BYVAL vhMax AS LONG, _
                              BYVAL vhPos AS LONG, BYVAL vhBar AS LONG)

'some useful, custom-made messages, see manual above
%VLB_SETARRAY        = %WM_USER + 1
%VLB_REFRESH         = %WM_USER + 2
%VLB_GETSELECTED     = %WM_USER + 3
%VLB_SETSELECTED     = %WM_USER + 4
%VLB_GETTOPLINE      = %WM_USER + 5
%VLB_SETTOPLINE      = %WM_USER + 6
%VLB_SIZEHANDLER     = %WM_USER + 7
%VLB_GETCURSEL       = %WM_USER + 8
%VLB_SETCURSEL       = %WM_USER + 9

%VLBN_RETURN     = %WM_USER + 20
%VLBN_SPACE      = %WM_USER + 21
%VLBN_DELETE     = %WM_USER + 22

TYPE vListData                   'Type variable to hold private ListBox data
  hParent             AS DWORD      'Can be freely customized to meet extended needs
  hInst               AS DWORD
  id                  AS DWORD
  cyChar              AS LONG
  wMaxHeight          AS LONG
  wFirstLine          AS LONG
  wLastLine           AS LONG
  tLineCount          AS LONG
  charWidth           AS LONG
  wMaxWidth           AS LONG
  iHscrollMax         AS LONG
  iHscrollPos         AS LONG
  xPos                AS LONG
  hBit                AS DWORD
  memDC               AS DWORD
  SelLine             AS LONG
  hFont               AS DWORD
  st                  AS STRING PTR
END TYPE

'********************************************************************
' Initialize and create the Virtual ListBox control
'--------------------------------------------------------------------
' hWnd             = Parent's (dialog's) handle
' CtrlId           = Control's Id
' vLeft, vTop      = position of upper left corner
' vWidth, vHeight  = control's width and height
' DDTstyle, %TRUE  = use DDT to create the control  (dialog units)
'           %FALSE = use CreateWindowEx to create it.  (pixels)
' Reason for this is that they use different coordinate systems
' and some prefer DDT, others SDK.. :-)
'********************************************************************
FUNCTION CreateVirtualList(BYVAL hParent  AS DWORD, BYVAL CtrlId  AS LONG, _
                           BYVAL vLeft    AS LONG,  BYVAL vTop    AS LONG, _
                           BYVAL vWidth   AS LONG,  BYVAL vHeight AS LONG, _
                           BYVAL DDTstyle AS LONG)  AS DWORD

  LOCAL hVList AS DWORD, wc AS WNDCLASSEX, szClassName AS ASCIIZ * 20

  szClassName      = "PBVLIST32"
  wc.cbSize        = SIZEOF(wc)
  wc.style         = %CS_DBLCLKS
  wc.lpfnWndProc   = CODEPTR(VListProc)
  wc.cbWndExtra    = 4  '4 extra bytes for pointer to UDT
  wc.hInstance     = GetWindowLong(hParent, %GWL_HINSTANCE)
  wc.hCursor       = LoadCursor(%NULL, BYVAL %IDC_ARROW)
  wc.lpszClassName = VARPTR(szClassName)

  RegisterClassEx wc

  IF DDTstyle THEN
     CONTROL ADD "PBVLIST32", hParent, CtrlId, "", vLeft, vTop, vWidth, vHeight, _
                 %WS_VISIBLE OR %WS_CHILD OR %WS_VSCROLL OR %WS_TABSTOP, %WS_EX_CLIENTEDGE
     CONTROL HANDLE hParent, CtrlId TO hVList
  ELSE
     hVList = CreateWindowEx(%WS_EX_CLIENTEDGE, "PBVLIST32", BYVAL 0, %WS_VISIBLE OR _
                             %WS_CHILD OR %WS_VSCROLL OR %WS_TABSTOP, _
                             vLeft, vTop, vWidth, vHeight, _
                             hParent, CtrlId, wc.hInstance, BYVAL 0)
  END IF

  FUNCTION = hVList 'return the handle
END FUNCTION

'********************************************************************
' Set Scroll bars
'********************************************************************
SUB SetPBVLScrollData(BYVAL hwnd AS DWORD, BYVAL vhPage AS LONG, BYVAL vhMax AS LONG, _
                      BYVAL vhPos AS LONG, BYVAL vhBar AS LONG)
  LOCAL si AS SCROLLINFO

  si.cbSize = SIZEOF(si)
  si.fMask  = %SIF_ALL OR %SIF_DISABLENOSCROLL
  si.nMin   = 0
  si.nMax   = MAX&(0, vhMax)
  si.nPage  = vhPage
  si.nPos   = vhPos
  SetScrollInfo hwnd, vhBar, si, -1

END SUB

'********************************************************************
' Main control procedure
'********************************************************************
FUNCTION VListProc (BYVAL hWnd AS DWORD, BYVAL wMsg AS DWORD, _
                    BYVAL wParam AS LONG, BYVAL lParam AS LONG) AS LONG

  'Note: v is declared as local, but stored globally via LocalAlloc and
  '      a pointer to the returned handle is stored in the extra bytes
  '      of the control's private window class (in cbWndExtra), so each
  '      control still can hold its own private data.

  LOCAL v AS VlistData PTR
  IF wMsg <> %WM_CREATE THEN v = GetWindowLong(hWnd, 0)

  SELECT CASE wMsg
     CASE %WM_CREATE  'Allocate storage for the vListData structure.
        LOCAL tm   AS TEXTMETRIC, ps AS PAINTSTRUCT,_
              si   AS SCROLLINFO, lp AS POINTAPI, _
              rc   AS RECT, wRect AS RECT, _
              hdc  AS DWORD, hPen AS DWORD, hBrush AS DWORD, _
              tSel AS LONG, bkBrush AS DWORD, hBrushSel AS DWORD, _
              y    AS LONG, i AS LONG, iVscrollInc AS LONG, hScrlInc AS LONG

        v = HeapAlloc(GetProcessHeap(), %HEAP_ZERO_MEMORY, SIZEOF(@v))
        IF v THEN
           SetWindowLong hWnd, 0, v  'Store the pointer for later use
        ELSE
           MessageBox hWnd, "Could not allocate memory for Virtual Listbox" & CHR$(0), _
                              "PBVLIST32 message" & CHR$(0), %MB_OK
           FUNCTION = -1 : EXIT FUNCTION  'Return -1 to break the action
        END IF

        @v.hParent = GetParent(hWnd)
        @v.id      = GetWindowLong(hWnd, %GWL_ID)
        @v.hInst   = GetWindowLong(hWnd, %GWL_HINSTANCE)
        @v.SelLine = -1 : @v.tLineCount = -1
        SendMessage hWnd, %VLB_SIZEHANDLER, 0, 0

     CASE %WM_DESTROY
        IF v THEN
           IF @v.hbit  THEN DeleteObject SelectObject(@v.memDC, @v.hbit)
           IF @v.memDC THEN DeleteDC @v.memDC
           CALL HeapFree(GetProcessHeap(), 0, BYVAL v)
        END IF

     CASE %WM_SETFONT
        IF wParam <> @v.hFont THEN
           @v.hFont = wParam
           SelectObject @v.memDC, @v.hFont
           CALL GetTextMetrics (@v.memDC, tm)              'Get font parameters
           @v.cyChar = tm.tmHeight + tm.tmExternalLeading  'Line spacing
           @v.charWidth  = tm.tmAveCharWidth               'Average character width
           CALL GetWindowRect(hWnd, wRect)                 'Adjust height to avoid partial lines
           @v.wMaxHeight = (wRect.nBottom - wRect.nTop) \ @v.cyChar
           @v.wMaxWidth = MAX&(1, ((wRect.nRight - wRect.nLeft) / 2) / (@v.charWidth - 1) ) 'Get window size in characters
           SetWindowPos hWnd, 0, 0, 0, wRect.nRight - wRect.nLeft, _
                              @v.wMaxHeight * @v.cyChar + 4, _
                              %SWP_NOMOVE OR %SWP_NOZORDER

           SendMessage hWnd, %VLB_SIZEHANDLER, 0, 0
        END IF

     CASE %WM_SIZE
        IF HIWRD(lParam) THEN
           @v.wMaxHeight = HIWRD(lParam) / @v.cyChar
           SendMessage hWnd, %VLB_SIZEHANDLER, 0, 0
        END IF

     CASE %VLB_SIZEHANDLER  'create a virtual window that fits current window size
        IF @v.hbit  THEN DeleteObject SelectObject(@v.memDC, @v.hbit)
        IF @v.memDC THEN DeleteDC @v.memDC

        hDC = GetDC(hWnd)
           CALL GetClientRect(hWnd, wRect)
           @v.memDC = CreateCompatibleDC(hDC)
           @v.hbit  = CreateCompatibleBitmap(hDC, wRect.nRight, wRect.nBottom)
           @v.hbit = SelectObject(@v.memDC, @v.hbit)

           hbrush   = GetStockObject(%WHITE_BRUSH)

           IF hbrush THEN SelectObject @v.memDC, hbrush
           IF @v.hFont = 0 THEN @v.hFont = GetStockObject(%ANSI_VAR_FONT)
           IF @v.hFont THEN SelectObject @v.memDC, @v.hFont
           PatBlt @v.memDC, 0, 0, wRect.nRight, wRect.nBottom, %PATCOPY

           CALL GetTextMetrics (@v.memDC, tm)              'Get font parameters
           @v.cyChar = tm.tmHeight + tm.tmExternalLeading  'Line spacing
           @v.charWidth = tm.tmAveCharWidth                'Average character width
        CALL ReleaseDC (hWnd, hdc)

        @v.wMaxHeight = wRect.nBottom / @v.cyChar
        @v.wMaxWidth = MAX&(1, (wRect.nRight / 2) / (@v.charWidth - 1) ) 'Get window size in characters

        '@v.iHscrollMax = MAX%(0, 1024 - @v.wMaxWidth) '1024, max number of characters..
        @v.iHscrollPos = MIN%(@v.iHscrollPos, @v.iHscrollMax)
        @v.xPos = @v.charWidth * (- @v.iHscrollPos)
        SetPBVLScrollData hwnd, @v.wMaxWidth, @v.iHscrollMax, @v.iHscrollPos, %SB_HORZ

        @v.wMaxHeight = wRect.nBottom / @v.cyChar
        SetPBVLScrollData hwnd, @v.wMaxHeight, MAX&(0, @v.tLineCount), @v.wFirstLine, %SB_VERT
        InvalidateRect hWnd, BYVAL %NULL, 0 : UpdateWindow hwnd

     CASE %WM_VSCROLL
        IF @v.tLineCount < 0 THEN EXIT SELECT

        SELECT CASE LOWRD(wParam)
           CASE %SB_TOP        : iVscrollInc = -@v.wFirstLine
           CASE %SB_BOTTOM     : iVscrollInc =  @v.tLineCount - @v.wMaxHeight
           CASE %SB_LINEUP     : iVscrollInc = -1
           CASE %SB_LINEDOWN   : IF @v.wFirstLine < @v.tLineCount - @v.wMaxHeight + 1 THEN iVscrollInc =  1
           CASE %SB_PAGEUP     : iVscrollInc = MIN&(-1, -@v.wMaxHeight)
           CASE %SB_PAGEDOWN   : iVscrollInc = MAX&(1,   @v.wMaxHeight)
           CASE %SB_THUMBTRACK ' getScrollInfo enables 32-bit scroll positions
              si.cbSize = SIZEOF(si)
              si.fMask  = %SIF_TRACKPOS
              GetScrollInfo hWnd, %SB_VERT, si
              iVscrollInc = si.nTrackPos - @v.wFirstLine

           CASE ELSE : iVscrollInc = 0
        END SELECT

        iVscrollInc = MAX&(-@v.wFirstLine, MIN&(iVscrollInc, @v.tLineCount - @v.wMaxHeight + 1))

        IF iVscrollInc <> 0 AND @v.wFirstLine <= @v.tLineCount - @v.wMaxHeight + 1 THEN
           @v.wFirstLine = MIN&(@v.wFirstLine + iVscrollInc, @v.tLineCount - @v.wMaxHeight + 1)
           SetPBVLScrollData hwnd, @v.wMaxHeight, MAX&(0, @v.tLineCount), @v.wFirstLine, %SB_VERT
           InvalidateRect hWnd, BYVAL %NULL, 0 : UpdateWindow hwnd
        END IF

     CASE %WM_HSCROLL
        SELECT CASE LOWRD(wParam)
           CASE %SB_LINELEFT      : hScrlInc = -1
           CASE %SB_LINERIGHT     : hScrlInc =  1
           CASE %SB_PAGELEFT      : hScrlInc = -@v.wMaxWidth
           CASE %SB_PAGERIGHT     : hScrlInc =  @v.wMaxWidth
           CASE %SB_THUMBTRACK    : hScrlInc = HIWRD(wParam) - @v.iHscrollPos
           CASE ELSE              : hScrlInc = 0
        END SELECT

        hScrlInc = MAX&(-@v.iHscrollPos, MIN&(hScrlInc, @v.iHscrollMax - @v.iHscrollPos))
        IF hScrlInc <> 0 THEN
           @v.iHscrollPos = @v.iHscrollPos + hScrlInc
           @v.xPos = @v.charWidth * (- @v.iHscrollPos)
           SetPBVLScrollData hwnd, @v.wMaxWidth, @v.iHscrollMax, @v.iHscrollPos, %SB_HORZ
           InvalidateRect hwnd, BYVAL %NULL, 0 : UpdateWindow hwnd
        END IF

     CASE %WM_GETDLGCODE ' Ensure that the control processes all keys (except tab and escape) by itself
        LOCAL iMsg AS LONG, pMsg AS TAGMSG PTR
        pMsg = lParam
        IF pMsg > 0 THEN iMsg = @pMsg.Message
        IF iMsg = %WM_KEYDOWN OR iMsg = %WM_CHAR THEN
           SELECT CASE @pMsg.wParam
              CASE %VK_TAB, %VK_ESCAPE   'let system handle these
                 FUNCTION = DefWindowProc(hWnd, wMsg, wParam, lParam)
              CASE ELSE
                 FUNCTION = %DLGC_WANTALLKEYS
           END SELECT
        END IF

     CASE %WM_CHAR                              ' mimic standard listbox's search on keypress
        FOR i = @v.SelLine + 1 TO @v.tLineCount ' look for next item that starts like pressed key
           IF wParam = ASC(@v.@st[i]) THEN      ' on success, set sel and exit
              SendMessage hWnd, %VLB_SETCURSEL, i, 1
              EXIT FUNCTION
           END IF
        NEXT

        'if no success and SelLine is > first item, scan from array start to SelLine - 1
        IF @v.SelLine > 0 THEN 
           FOR i = 0 TO @v.SelLine - 1
              IF wParam = ASC(@v.@st[i]) THEN
                 SendMessage hWnd, %VLB_SETCURSEL, i, 1
                 EXIT FUNCTION
              END IF
           NEXT
        END IF

     CASE %WM_KEYDOWN
        IF @v.tLineCount < 0 AND wParam <> %VK_TAB THEN EXIT SELECT

        SELECT CASE (wParam)
           CASE %VK_UP, %VK_LEFT
              IF @v.SelLine = -1 THEN @v.SelLine = @v.wFirstLine + 1
              @v.SelLine = MAX&(@v.SelLine - 1, 0)
              IF @v.SelLine < @v.wFirstLine OR @v.SelLine > @v.wLastLine THEN
                 @v.wFirstLine = MIN&(@v.SelLine, @v.tLineCount - @v.wMaxHeight + 1)
                 SetPBVLScrollData hwnd, @v.wMaxHeight, MAX&(0, @v.tLineCount), @v.wFirstLine, %SB_VERT
              END IF
              InvalidateRect hWnd, BYVAL %NULL, 0 : UpdateWindow hwnd

           CASE %VK_DOWN, %VK_RIGHT
              IF @v.SelLine = -1 THEN @v.SelLine = @v.wFirstLine - 1
              @v.SelLine = MIN&(@v.SelLine + 1, @v.tLineCount)
              IF @v.SelLine > @v.wLastLine OR @v.SelLine < @v.wFirstLine THEN
                 @v.wFirstLine = MIN&(@v.SelLine, @v.tLineCount) - @v.wMaxHeight + 1
                 SetPBVLScrollData hwnd, @v.wMaxHeight, MAX&(0, @v.tLineCount), @v.wFirstLine, %SB_VERT
              END IF
              InvalidateRect hWnd, BYVAL %NULL, 0 : UpdateWindow hwnd

           CASE %VK_PRIOR 'PgUp
              IF @v.SelLine = @v.wFirstLine THEN
                 @v.SelLine = MAX&(@v.SelLine - @v.wMaxHeight + 1, 0)
              ELSE
                 @v.SelLine = @v.wFirstLine
              END IF

              IF @v.SelLine < @v.wFirstLine + 1 THEN
                 @v.wFirstLine = @v.SelLine
                 CALL SetScrollPos (hWnd, %SB_VERT, @v.wFirstLine, %TRUE)
              END IF
              InvalidateRect hWnd, BYVAL %NULL, 0 : UpdateWindow hwnd

           CASE %VK_NEXT  'PgDn
              IF @v.SelLine = @v.wLastLine THEN
                 @v.SelLine = MIN&(@v.SelLine + @v.wMaxHeight - 1, @v.tLineCount)
              ELSE
                 @v.SelLine = @v.wLastLine
              END IF

              IF @v.SelLine > @v.wLastLine THEN
                 @v.wFirstLine = MIN&(@v.SelLine, @v.tLineCount) - @v.wMaxHeight + 1
                 CALL SetScrollPos (hWnd, %SB_VERT, @v.wFirstLine, %TRUE)
              END IF
              InvalidateRect hWnd, BYVAL %NULL, 0 : UpdateWindow hwnd

           CASE %VK_SPACE  'Space bar pressed
              SendMessage @v.hParent, %WM_COMMAND, MAKLNG(@v.id, %VLBN_SPACE), hwnd

           CASE %VK_RETURN 'Enter key pressed
              SendMessage @v.hParent, %WM_COMMAND, MAKLNG(@v.id, %VLBN_RETURN), hwnd
              SendMessage @v.hParent, %WM_COMMAND, MAKLNG(@v.id, %IDOK), hwnd

           CASE %VK_DELETE 'Delete key pressed
              SendMessage @v.hParent, %WM_COMMAND, MAKLNG(@v.id, %VLBN_DELETE), hwnd

           CASE %VK_HOME
              @v.SelLine = 0 : @v.wFirstLine = 0
              CALL SetScrollPos (hWnd, %SB_VERT, @v.wFirstLine, %TRUE)
              InvalidateRect hWnd, BYVAL %NULL, 0 : UpdateWindow hwnd

           CASE %VK_END
              @v.SelLine = @v.tLineCount
              @v.wFirstLine = @v.tLineCount - @v.wMaxHeight + 1
              CALL SetScrollPos (hWnd, %SB_VERT, @v.wFirstLine, %TRUE)
              InvalidateRect hWnd, BYVAL %NULL, 0 : UpdateWindow hwnd
        END SELECT

     CASE %WM_MOUSEMOVE
        IF @v.tLineCount < 0 THEN EXIT SELECT
        IF wParam = %MK_LBUTTON THEN
           GetCursorPos lp : GetWindowRect hWnd, wRect
           IF lp.y < wRect.nTop + 3 THEN
              tSel = @v.wFirstLine - ((wRect.nTop - lp.y) \ @v.cyChar + 1)
           ELSE
              tSel = @v.wFirstLine + HIWRD(lParam) \ @v.cyChar
           END IF

           IF tSel < 0 THEN tSel = 0
           IF tSel > @v.tLineCount THEN tSel = @v.tLineCount
           IF @v.SelLine = tSel THEN EXIT SELECT 'no need to repeat ourselves..

           IF tSel > @v.wLastLine THEN @v.wFirstLine = tSel - @v.wMaxHeight + 1
           IF tSel < @v.wFirstLine THEN @v.wFirstLine = tSel
           @v.SelLine = tSel

           IF @v.tLineCount > @v.wMaxHeight - 1 THEN
              SetPBVLScrollData hwnd, @v.wMaxHeight, MAX&(0, @v.tLineCount), @v.wFirstLine, %SB_VERT
           END IF

           InvalidateRect hWnd, BYVAL %NULL, 0 : UpdateWindow hwnd
        END IF

     CASE %WM_LBUTTONDBLCLK
        IF @v.tLineCount < 0 THEN EXIT SELECT
        @v.SelLine = MAX&(0, @v.wFirstLine + HIWRD(lParam) \ @v.cyChar)
        IF @v.SelLine <= @v.tLineCount THEN
           SendMessage @v.hParent, %WM_COMMAND, MAKLNG(@v.id, %LBN_DBLCLK), hwnd
        END IF

     CASE %WM_LBUTTONDOWN
        IF @v.tLineCount < 0 THEN EXIT SELECT
        SetCapture hWnd
        @v.SelLine = @v.wFirstLine + HIWRD(lParam) \ @v.cyChar
        IF @v.SelLine < 0 THEN @v.SelLine = 0
        IF @v.SelLine > @v.tLineCount THEN @v.SelLine = @v.tLineCount
        SetFocus hWnd
        InvalidateRect hWnd, BYVAL %NULL, 0 : UpdateWindow hwnd

     CASE %WM_LBUTTONUP
        IF @v.tLineCount < 0 THEN EXIT SELECT
        ReleaseCapture

     CASE %WM_KILLFOCUS, %WM_SETFOCUS ' Must process these if line is selected
        IF @v.SelLine > - 1 THEN
           InvalidateRect hWnd, BYVAL %NULL, 0 : UpdateWindow hwnd
        ELSE
           GetClientRect hWnd, wRect
           rc.nLeft = 0 : rc.nTop = 0 : rc.nRight = wRect.nright : rc.nBottom = @v.cyChar
           hdc = GetDc(hWnd)
             DrawFocusRect hDC, rc
           ReleaseDc hWnd, hdc
        END IF
        IF wMsg = %WM_KILLFOCUS THEN 'Send notifications to parent
           SendMessage @v.hParent, %WM_COMMAND, MAKLNG(@v.id, %LBN_KILLFOCUS), hwnd
        ELSE
           SendMessage @v.hParent, %WM_COMMAND, MAKLNG(@v.id, %LBN_SETFOCUS), hwnd
        END IF

     CASE %WM_PAINT 'Draw only the list items that needs to be shown
        'Send %WM_CTLCOLORLISTBOX to parent, to get eventual brush for bg color
        bkBrush = SendMessage(@v.hParent, %WM_CTLCOLORLISTBOX, @v.memDC, hwnd)
        IF bkBrush THEN SelectObject @v.memDC, bkBrush

        hdc = BeginPaint(hWnd, ps)
           CALL GetClientRect(hWnd, wRect)

           @v.wFirstLine = MAX&(0, @v.wFirstLine)
           @v.wLastLine  = MIN&(@v.tLineCount, @v.wFirstLine + @v.wMaxHeight - 1)

           PatBlt @v.memDC, 0, 0, wRect.nRight, wRect.nBottom, %PATCOPY

           IF @v.tLineCount > -1 THEN ' DRAW TEXT (ARRAY)
              FOR i = @v.wFirstLine TO @v.wLastLine
                 y = @v.cyChar * (-@v.wFirstLine + i)
                 TabbedTextOut @v.memDC, @v.xPos, y, _
                               BYVAL STRPTR(@v.@st[i]), LEN(@v.@st[i]), _
                               0, BYVAL %NULL, 0
              NEXT
           END IF

           'DRAW SELECTION
           IF @v.SelLine >= @v.wFirstLine THEN
              IF @v.SelLine > @v.tLineCount THEN @v.SelLine = @v.tLineCount
              rc.nLeft = 0 : rc.nTop = @v.cyChar * (@v.SelLine - @v.wFirstLine)
              rc.nRight = wRect.nright : rc.nBottom = rc.nTop + @v.cyChar

              '=== Draw Selection rectangle, plus focus rectangle ===
              IF GetFocus = hWnd THEN
                 hBrushSel = GetSysColorBrush(%COLOR_HIGHLIGHT)
              ELSE
                 hBrushSel = GetSysColorBrush(%COLOR_INACTIVECAPTION)
              END IF
              hBrushSel = SelectObject(@v.memDC, hBrushSel)
              hPen = CreatePen(%PS_DOT, 1, GetSysColor(%COLOR_HIGHLIGHTTEXT))
              hPen = SelectObject(@v.memDC, hPen)

              CALL SetROP2(@v.memDC, %R2_MERGEPENNOT)
              CALL Rectangle(@v.memDC, rc.nLeft - 1, rc.nTop - 1, rc.nRight + 1, rc.nBottom + 1)
              IF GetFocus() = hWnd THEN CALL DrawFocusRect(@v.memDC, rc)

              SelectObject @v.memDC, hBrushSel
              DeleteObject SelectObject(@v.memDC, hPen)
              SendMessage @v.hParent, %WM_COMMAND, MAKLNG(@v.id, %LBN_SELCHANGE), hwnd

           ELSEIF @v.SelLine = -1 AND GetFocus = hWnd THEN
              rc.nLeft = 0 : rc.nTop = 0 : rc.nRight = wRect.nright : rc.nBottom = @v.cyChar
              DrawFocusRect @v.memDC, rc
           END IF

           BitBlt hDC, wRect.ntop, wRect.nleft, wRect.nright, wRect.nbottom, @v.memDC, 0, 0, %SRCCOPY
        CALL EndPaint(hWnd, ps)

     CASE %WM_ERASEBKGND 'Prevent erasing background
        FUNCTION = 1

     CASE %LB_SETHORIZONTALEXTENT
        CALL GetClientRect(hWnd, wRect)
        @v.iHscrollMax = MAX%(0, wParam - wRect.nRight)
        SetPBVLScrollData hwnd, @v.wMaxWidth, @v.iHscrollMax, @v.iHscrollPos, %SB_HORZ

     CASE %VLB_SETARRAY
        @v.st = wParam : @v.SelLine = -1 : @v.tLineCount = lParam
        @v.wFirstLine = 0

     CASE %VLB_REFRESH
        SetPBVLScrollData hwnd, @v.wMaxHeight, MAX&(0, @v.tLineCount), @v.wFirstLine, %SB_VERT
        InvalidateRect hWnd, BYVAL %NULL, 0 : UpdateWindow hwnd

     CASE %VLB_GETSELECTED
        FUNCTION = @v.SelLine

     CASE %VLB_GETCURSEL
        FUNCTION = @v.SelLine

     CASE %VLB_SETSELECTED
        IF wParam < 0 THEN
           @v.SelLine = -1
        ELSE
           @v.SelLine = MIN&(wParam, @v.tLineCount)
        END IF
        SetPBVLScrollData hwnd, @v.wMaxHeight, MAX&(0, @v.tLineCount), @v.wFirstLine, %SB_VERT
        IF lParam THEN 'Refresh control
           InvalidateRect hWnd, BYVAL %NULL, 0 : UpdateWindow hwnd
        END IF
        FUNCTION = @v.SelLine

     CASE %VLB_SETCURSEL
        IF wParam < 0 THEN
           @v.SelLine = -1
        ELSE
           @v.SelLine = MIN&(wParam, @v.tLineCount)
           IF @v.SelLine < @v.wFirstLine THEN
              @v.wFirstLine = @v.SelLine
           ELSEIF @v.SelLine > @v.wFirstLine + @v.wMaxHeight - 1 THEN
              @v.wFirstLine = MIN&(@v.SelLine - @v.wMaxHeight + 1, @v.tLineCount - @v.wMaxHeight + 1)
           END IF
        END IF

        SetPBVLScrollData hwnd, @v.wMaxHeight, MAX&(0, @v.tLineCount), @v.wFirstLine, %SB_VERT
        InvalidateRect hWnd, BYVAL %NULL, 0 : UpdateWindow hwnd
        FUNCTION = @v.SelLine

     CASE %VLB_GETTOPLINE
        FUNCTION = @v.wFirstLine

     CASE %VLB_SETTOPLINE
        IF wParam < 0 THEN
           @v.wFirstLine = 0
        ELSE
           @v.wFirstLine = MIN&(wParam, @v.tLineCount - @v.wMaxHeight + 1)
        END IF
        SetPBVLScrollData hwnd, @v.wMaxHeight, MAX&(0, @v.tLineCount), @v.wFirstLine, %SB_VERT
        IF lParam THEN 'Refresh control
           InvalidateRect hWnd, BYVAL %NULL, 0 : UpdateWindow hwnd
        END IF
        FUNCTION = @v.wFirstLine

     CASE ELSE  ' Process all other messages
        FUNCTION = DefWindowProc(hWnd, wMsg, wParam, lParam)

  END SELECT

END FUNCTION
